'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var _objectSpread = require('@babel/runtime/helpers/objectSpread2');
var _defineProperty = require('@babel/runtime/helpers/defineProperty');
var core = require('@remirror/core');
var _createForOfIteratorHelper = require('@babel/runtime/helpers/createForOfIteratorHelper');
var _classPrivateFieldGet = require('@babel/runtime/helpers/classPrivateFieldGet');
var _classPrivateFieldSet = require('@babel/runtime/helpers/classPrivateFieldSet');
var CodeMirror = require('codemirror');
var commands = require('@remirror/pm/commands');
var history = require('@remirror/pm/history');
var state = require('@remirror/pm/state');
require('codemirror/mode/meta');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var _objectSpread__default = /*#__PURE__*/_interopDefault(_objectSpread);
var _defineProperty__default = /*#__PURE__*/_interopDefault(_defineProperty);
var _createForOfIteratorHelper__default = /*#__PURE__*/_interopDefault(_createForOfIteratorHelper);
var _classPrivateFieldGet__default = /*#__PURE__*/_interopDefault(_classPrivateFieldGet);
var _classPrivateFieldSet__default = /*#__PURE__*/_interopDefault(_classPrivateFieldSet);
var CodeMirror__default = /*#__PURE__*/_interopDefault(CodeMirror);

/**
 * Handling cursor motion from the outer to the inner editor must be done with a
 * keymap on the outer editor. The `arrowHandler` function uses the
 * `endOfTextblock` method to determine, in a bidi-text-aware way, whether the
 * cursor is at the end of a given textblock. If it is, and the next block is a
 * code block, the selection is moved into it.
 *
 * Adapted from https://prosemirror.net/examples/codemirror/
 */
function arrowHandler(dir) {
  return (_ref) => {
    var dispatch = _ref.dispatch,
        view = _ref.view,
        tr = _ref.tr;

    if (!view) {
      return false;
    }

    if (!(tr.selection.empty && view.endOfTextblock(dir))) {
      return false;
    }

    var side = dir === 'left' || dir === 'up' ? -1 : 1;
    var $head = tr.selection.$head;
    var nextPos = state.Selection.near(tr.doc.resolve(side > 0 ? $head.after() : $head.before()), side);

    if (nextPos.$head && nextPos.$head.parent.type.name === 'codeMirror') {
      dispatch === null || dispatch === void 0 ? void 0 : dispatch(tr.setSelection(nextPos));
      return true;
    }

    return false;
  };
}
/**
 * Updates the node attrs.
 *
 * This is used to update the language for the CodeMirror block.
 */

function updateNodeAttributes(type) {
  return attributes => (_ref2) => {
    var _ref2$state = _ref2.state,
        tr = _ref2$state.tr,
        selection = _ref2$state.selection,
        dispatch = _ref2.dispatch;
    var parent = core.findParentNodeOfType({
      types: type,
      selection
    });

    if (!parent || core.isEqual(attributes, parent.node.attrs)) {
      // Do nothing since the attrs are the same
      return false;
    }

    tr.setNodeMarkup(parent.pos, type, _objectSpread__default['default'](_objectSpread__default['default']({}, parent.node.attrs), attributes));

    if (dispatch) {
      dispatch(tr);
    }

    return true;
  };
}
/**
 * Parse a language string to CodeMirror's mode option.
 */

function parseLanguageToMode(language) {
  if (language) {
    var mime = CodeMirror__default['default'].findModeByName(language);
    return mime === null || mime === void 0 ? void 0 : mime.mode;
  }

  return null;
}

var _node = /*#__PURE__*/new WeakMap();

var _cm = /*#__PURE__*/new WeakMap();

var _schema = /*#__PURE__*/new WeakMap();

var _view = /*#__PURE__*/new WeakMap();

var _getPos = /*#__PURE__*/new WeakMap();

var _incomingChanges = /*#__PURE__*/new WeakMap();

var _updating = /*#__PURE__*/new WeakMap();

class CodeMirrorNodeView {
  /**
   * The dom element which wraps the CodeMirror editor element. This is directly
   * controlled by ProseMirror.
   */

  /**
   * The ProseMirror Node represented by this instance of the
   * [[`CodeMirrorNodeView`]].
   */

  /**
   * The CodeMirror editor instance being controlled by this NodeView.
   */

  /**
   * The schema for the current editor.
   */

  /**
   * The `EditorView` being used.
   */

  /**
   * A method for retrieving the current position from the editor.
   */

  /**
   * A flag used to identify whether there are any changes that are incoming. It
   * is used along with `this.#updating` to prevent cursorActivity changes from
   * being forwarded to the ProseMirror editor.
   */

  /**
   * This flag is used to avoid an update loop between the outer and inner
   * editor.
   *
   * - `true` when the CodeMirror editor is being updated right now.
   */
  constructor(node, view, getPos, config) {
    _node.set(this, {
      writable: true,
      value: void 0
    });

    _cm.set(this, {
      writable: true,
      value: void 0
    });

    _schema.set(this, {
      writable: true,
      value: void 0
    });

    _view.set(this, {
      writable: true,
      value: void 0
    });

    _getPos.set(this, {
      writable: true,
      value: void 0
    });

    _incomingChanges.set(this, {
      writable: true,
      value: false
    });

    _updating.set(this, {
      writable: true,
      value: false
    });

    // Store for later
    _classPrivateFieldSet__default['default'](this, _node, node);

    _classPrivateFieldSet__default['default'](this, _schema, node.type.schema);

    _classPrivateFieldSet__default['default'](this, _view, view);

    _classPrivateFieldSet__default['default'](this, _getPos, getPos); // Create a CodeMirror instance


    _classPrivateFieldSet__default['default'](this, _cm, CodeMirror__default['default'](null, _objectSpread__default['default']({
      value: _classPrivateFieldGet__default['default'](this, _node).textContent,
      extraKeys: this.codeMirrorKeymap()
    }, config))); // The editor's outer node is our DOM representation


    this.dom = _classPrivateFieldGet__default['default'](this, _cm).getWrapperElement();
    this.setupCodeMirrorHandlers();
  }
  /**
   * Create the event listeners for managing updates from CodeMirror.
   */


  setupCodeMirrorHandlers() {
    // CodeMirror needs to be in the DOM to properly initialize, so schedule it
    // to update itself.
    setTimeout(() => _classPrivateFieldGet__default['default'](this, _cm).refresh(), 20); //

    _classPrivateFieldSet__default['default'](this, _updating, false); // Track whether changes are have been made but not yet propagated.


    _classPrivateFieldGet__default['default'](this, _cm).on('beforeChange', () => _classPrivateFieldSet__default['default'](this, _incomingChanges, true)); // Propagate updates from the code editor to ProseMirror.


    _classPrivateFieldGet__default['default'](this, _cm).on('cursorActivity', () => {
      if (!_classPrivateFieldGet__default['default'](this, _updating) && !_classPrivateFieldGet__default['default'](this, _incomingChanges)) {
        this.forwardSelection();
      }
    }); // Listen to changes within the code editor.


    _classPrivateFieldGet__default['default'](this, _cm).on('changes', () => {
      if (!_classPrivateFieldGet__default['default'](this, _updating)) {
        this.valueChanged();
        this.forwardSelection();
      }

      _classPrivateFieldSet__default['default'](this, _incomingChanges, false);
    });

    _classPrivateFieldGet__default['default'](this, _cm).on('focus', () => this.forwardSelection());
  }
  /**
   * When the code editor is focused, we can keep the selection of the outer
   * editor synchronized with the inner one, so that any commands executed on
   * the outer editor see an accurate selection.
   */


  forwardSelection() {
    if (!_classPrivateFieldGet__default['default'](this, _cm).hasFocus()) {
      return;
    }

    var state = _classPrivateFieldGet__default['default'](this, _view).state;

    var selection = this.asProseMirrorSelection(state.doc);

    if (!selection.eq(state.selection)) {
      _classPrivateFieldGet__default['default'](this, _view).dispatch(state.tr.setSelection(selection));
    }
  }
  /**
   * This helper function translates from a CodeMirror selection to a
   * ProseMirror selection. Because CodeMirror uses a line/column based indexing
   * system, `indexFromPos` is used to convert to an actual character index.
   */


  asProseMirrorSelection(doc) {
    var offset = _classPrivateFieldGet__default['default'](this, _getPos).call(this) + 1;
    var anchor = _classPrivateFieldGet__default['default'](this, _cm).indexFromPos(_classPrivateFieldGet__default['default'](this, _cm).getCursor('anchor')) + offset;
    var head = _classPrivateFieldGet__default['default'](this, _cm).indexFromPos(_classPrivateFieldGet__default['default'](this, _cm).getCursor('head')) + offset;
    return state.TextSelection.create(doc, anchor, head);
  }
  /**
   * Selections are also synchronized the other way, from ProseMirror to
   * CodeMirror, using the view's `setSelection` method.
   */


  setSelection(anchor, head) {
    _classPrivateFieldGet__default['default'](this, _cm).focus();

    _classPrivateFieldSet__default['default'](this, _updating, true);

    _classPrivateFieldGet__default['default'](this, _cm).setSelection(_classPrivateFieldGet__default['default'](this, _cm).posFromIndex(anchor), _classPrivateFieldGet__default['default'](this, _cm).posFromIndex(head));

    _classPrivateFieldSet__default['default'](this, _updating, false);
  }
  /**
   * When the actual content of the code editor is changed, the event handler
   * registered in the node view's constructor calls this method. It'll compare
   * the code block node's current value to the value in the editor, and
   * dispatch a transaction if there is a difference.
   */


  valueChanged() {
    var change = computeTextChange(_classPrivateFieldGet__default['default'](this, _node).textContent, _classPrivateFieldGet__default['default'](this, _cm).getValue()); // Return early when nothings has changed.

    if (!change) {
      return;
    }

    var start = _classPrivateFieldGet__default['default'](this, _getPos).call(this) + 1;

    var tr = _classPrivateFieldGet__default['default'](this, _view).state.tr.replaceWith(start + change.from, start + change.to, change.text ? _classPrivateFieldGet__default['default'](this, _schema).text(change.text) : []);

    _classPrivateFieldGet__default['default'](this, _view).dispatch(tr);
  }
  /**
   * A somewhat tricky aspect of nesting editor like this is handling cursor
   * motion across the edges of the inner editor. This node view will have to
   * take care of allowing the user to move the selection out of the code
   * editor. For that purpose, it binds the arrow keys to handlers that check if
   * further motion would ‘escape’ the editor, and if so, return the selection
   * and focus to the outer editor.
   *
   * The keymap also binds keys for undo and redo, which the outer editor will
   * handle, and for ctrl-enter, which, in ProseMirror's base keymap, creates a
   * new paragraph after a code block.
   */


  codeMirrorKeymap() {
    var view = _classPrivateFieldGet__default['default'](this, _view);

    var mod = /Mac/.test(navigator.platform) ? 'Cmd' : 'Ctrl';
    return CodeMirror__default['default'].normalizeKeyMap({
      Up: () => this.maybeEscape('line', -1),
      Left: () => this.maybeEscape('char', -1),
      Down: () => this.maybeEscape('line', 1),
      Right: () => this.maybeEscape('char', 1),
      'Ctrl-Enter': () => {
        if (commands.exitCode(view.state, view.dispatch)) {
          view.focus();
        }
      },
      ["".concat(mod, "-Z")]: () => history.undo(view.state, view.dispatch),
      ["Shift-".concat(mod, "-Z")]: () => history.redo(view.state, view.dispatch),
      ["".concat(mod, "-Y")]: () => history.redo(view.state, view.dispatch)
    });
  }

  maybeEscape(unit, dir) {
    var pos = _classPrivateFieldGet__default['default'](this, _cm).getCursor();

    if (_classPrivateFieldGet__default['default'](this, _cm).somethingSelected() || pos.line !== (dir < 0 ? _classPrivateFieldGet__default['default'](this, _cm).firstLine() : _classPrivateFieldGet__default['default'](this, _cm).lastLine()) || unit === 'char' && pos.ch !== (dir < 0 ? 0 : _classPrivateFieldGet__default['default'](this, _cm).getLine(pos.line).length)) {
      return CodeMirror__default['default'].Pass;
    }

    _classPrivateFieldGet__default['default'](this, _view).focus();

    var targetPos = _classPrivateFieldGet__default['default'](this, _getPos).call(this) + (dir < 0 ? 0 : _classPrivateFieldGet__default['default'](this, _node).nodeSize);
    var selection = state.Selection.near(_classPrivateFieldGet__default['default'](this, _view).state.doc.resolve(targetPos), dir);

    _classPrivateFieldGet__default['default'](this, _view).dispatch(_classPrivateFieldGet__default['default'](this, _view).state.tr.setSelection(selection).scrollIntoView());

    _classPrivateFieldGet__default['default'](this, _view).focus();

    return null;
  }
  /**
   * When an update comes in from the editor, for example because of an undo
   * action, we kind of have to do the inverse of what `valueChanged` did—check
   * for text changes, and if present, propagate them from the outer to the
   * inner editor.
   */


  update(node) {
    if (node.type !== _classPrivateFieldGet__default['default'](this, _node).type) {
      return false;
    }

    var textChange = computeTextChange(_classPrivateFieldGet__default['default'](this, _cm).getValue(), node.textContent);
    var attrsChange = computeAttrsChange(_classPrivateFieldGet__default['default'](this, _node).attrs, node.attrs);

    _classPrivateFieldSet__default['default'](this, _node, node);

    if (textChange || attrsChange) {
      _classPrivateFieldSet__default['default'](this, _updating, true);

      if (textChange) {
        _classPrivateFieldGet__default['default'](this, _cm).replaceRange(textChange.text, _classPrivateFieldGet__default['default'](this, _cm).posFromIndex(textChange.from), _classPrivateFieldGet__default['default'](this, _cm).posFromIndex(textChange.to));
      }

      if (attrsChange) {
        var _iterator = _createForOfIteratorHelper__default['default'](Object.keys(attrsChange)),
            _step;

        try {
          for (_iterator.s(); !(_step = _iterator.n()).done;) {
            var k = _step.value;

            _classPrivateFieldGet__default['default'](this, _cm).setOption(k, attrsChange[k]);
          }
        } catch (err) {
          _iterator.e(err);
        } finally {
          _iterator.f();
        }
      }

      _classPrivateFieldSet__default['default'](this, _updating, false);
    }

    return true;
  }

  selectNode() {
    _classPrivateFieldGet__default['default'](this, _cm).focus();
  }

  stopEvent() {
    return true;
  }

}
/**
 * `computeTextChange` compare two strings and find the minimal change between
 * them.
 *
 * It iterates from the start and end of the strings, until it hits a
 * difference, and returns an object giving the change's start, end, and
 * replacement text, or `null` if there was no change.
 */

function computeTextChange(previousText, currentText) {
  // Exit early if the strings are identical.
  if (previousText === currentText) {
    return null;
  } // Keep track of where the change starts.


  var from = 0; // Track the end position of relative to the original value.

  var to = previousText.length; // Track the end position relative the the current value.

  var currentTo = currentText.length; // Step forwards from the starting point until a changed value is encountered
  // and store the index of that changed value.

  while (from < to && previousText.charCodeAt(from) === currentText.charCodeAt(from)) {
    ++from;
  } // Step backwards from the end of the text values until a character which
  // doesn't match is encoutered. Store the index where the change happened in
  // both the `previousText` and the `currentText`.


  while (to > from && currentTo > from && previousText.charCodeAt(to - 1) === currentText.charCodeAt(currentTo - 1)) {
    to--;
    currentTo--;
  }

  return {
    from,
    to,
    text: currentText.slice(from, currentTo)
  };
}

function computeAttrsChange(oldAttrs, newAttrs) {
  var _oldAttrs$codeMirrorC, _newAttrs$codeMirrorC;

  var updated = false;
  var deltaConfig = {};
  var oldConfig = (_oldAttrs$codeMirrorC = oldAttrs.codeMirrorConfig) !== null && _oldAttrs$codeMirrorC !== void 0 ? _oldAttrs$codeMirrorC : {};
  var newConfig = (_newAttrs$codeMirrorC = newAttrs.codeMirrorConfig) !== null && _newAttrs$codeMirrorC !== void 0 ? _newAttrs$codeMirrorC : {};

  var _iterator2 = _createForOfIteratorHelper__default['default'](Object.keys(oldConfig)),
      _step2;

  try {
    for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
      var key = _step2.value;

      if (oldConfig[key] !== newConfig[key]) {
        deltaConfig[key] = newConfig[key];
        updated = true;
      }
    }
  } catch (err) {
    _iterator2.e(err);
  } finally {
    _iterator2.f();
  }

  if (oldAttrs.language !== newAttrs.language) {
    deltaConfig['mode'] = parseLanguageToMode(newAttrs.language);
    updated = true;
  }

  if (updated) {
    return deltaConfig;
  }

  return null;
}

var _dec, _class, _temp;
var CodeMirrorExtension = (_dec = /*#__PURE__*/core.extensionDecorator({
  defaultOptions: {
    defaultCodeMirrorConfig: null
  }
}), /*#__PURE__*/_dec(_class = (_temp = class CodeMirrorExtension extends core.NodeExtension {
  constructor() {
    super(...arguments);

    _defineProperty__default['default'](this, "tags", [core.ExtensionTag.BlockNode, core.ExtensionTag.Code]);
  }

  get name() {
    return 'codeMirror';
  }

  createNodeSpec(extra) {
    return {
      group: 'block',
      content: 'text*',
      marks: '',
      code: true,
      defining: true,
      attrs: _objectSpread__default['default'](_objectSpread__default['default']({}, extra.defaults()), {}, {
        codeMirrorConfig: {
          default: undefined
        },
        language: {
          default: undefined
        }
      }),
      parseDOM: [{
        tag: 'pre',
        getAttrs: node => core.isElementDomNode(node) ? extra.parse(node) : false
      }],

      toDOM() {
        return ['pre', ['code', 0]];
      },

      isolating: true
    };
  }

  createNodeViews() {
    return (node, view, getPos) => {
      var codeMirrorConfig = _objectSpread__default['default'](_objectSpread__default['default']({}, this.options.defaultCodeMirrorConfig), node.attrs.codeMirrorConfig);

      if (node.attrs.language) {
        var mode = parseLanguageToMode(node.attrs.language);

        if (mode) {
          codeMirrorConfig.mode = mode;
        }
      }

      return new CodeMirrorNodeView(node, view, getPos, codeMirrorConfig);
    };
  }

  createKeymap() {
    return {
      ArrowLeft: arrowHandler('left'),
      ArrowRight: arrowHandler('right'),
      ArrowUp: arrowHandler('up'),
      ArrowDown: arrowHandler('down')
    };
  }

  createCommands() {
    return {
      /**
       * Creates a CodeMirror block at the current position.
       *
       * ```ts
       * commands.createCodeMirror({ language: 'js' });
       * ```
       */
      createCodeMirror: attributes => {
        return core.setBlockType(this.type, attributes);
      },

      /**
       * Update the code block at the current position. Primarily this is used
       * to change the language.
       *
       * ```ts
       * if (commands.updateCodeMirror.isEnabled()) {
       *   commands.updateCodeMirror({ language: 'markdown' });
       * }
       * ```
       */
      updateCodeMirror: attributes => updateNodeAttributes(this.type)(attributes)
    };
  }

}, _temp)) || _class);

exports.CodeMirrorExtension = CodeMirrorExtension;
